/*
    mploader.cpp - main file for IDA loader for Newton MP 2K images
           Copyright (C) 2004
           Alexey Danilchenko

    This file is part of the mploader.
    Mploader is free software; you can redistribute it and/or
    modify it for whatever needs providing that you don't break
    any copyrights and laws.
    
    Mploader is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    
    Author does not hold the responsibility for any damages or 
    improper or illegal use of the mpdumper.
    
    This software uses parts of GNU libiberty library covered by 
    GPL version 2 license. 

*/

#include <string.h>
#include <stdio.h>

#include "demangle.h"
#include "mploader.h"

// IDA includes
#include <ida.hpp>
#include <fpro.h>
#include <idp.hpp>
#include <loader.hpp>
#include <name.hpp>
#include <bytes.hpp>
#include <offset.hpp>
#include <segment.hpp>
#include <srarea.hpp>
#include <fixup.hpp>
#include <entry.hpp>
#include <auto.hpp>
#include <diskio.hpp>
#include <kernwin.hpp>


// ------------------------ Global data ------------------------ 

// program option - do name truncation for the function names
static bool truncate_function_name = false;

// endian type
static EndianType endian = ET_BIG;

// SWI 0x11 operation encoded in big and little endians
static byte prog_exit_big_endian[] = {0xEF, 0x00, 0x00, 0x11};
static byte prog_exit_little_endian[] = {0x11, 0x00, 0x00, 0xEF};

// file data
static bool isMP2KUS = false;
static uint32 mp2kus_image_size = 9349820;    // in bytes

static AIFHeader header;
static Symbol* symbols = 0;
static byte * string_table = 0;
static byte * debug_data = 0;

static uint32 code_offset = 0;
static uint32 code_base = 0;
static uint32 code_length = 0;

static uint32 data_offset = 0;
static uint32 data_base = 0;
static uint32 data_length = 0;

static uint32 zero_data_base = 0;
static uint32 zero_data_length = 0;

static uint32 debug_offset = 0;
static uint32 debug_length = 0;

static uint32 mmu_base =   0x01A00000;
static uint32 mmu_length = 0x01D00000 - mmu_base;

static MemoryMapping mmu[] = {
    { 0x00002000, 0x01A00000 }, 
    { 0x00002080, 0x01A01080 }, 
    { 0x00002100, 0x01A02100 }, 
    { 0x00002180, 0x01A03180 }, 
    { 0x00002200, 0x01A04200 }, 
    { 0x00002280, 0x01A05280 }, 
    { 0x00002300, 0x01A06300 }, 
    { 0x00002380, 0x01A07380 }, 
    { 0x00002400, 0x01A08400 }, 
    { 0x00002480, 0x01A09480 }, 
    { 0x00002500, 0x01A0A500 }, 
    { 0x00002580, 0x01A0B580 }, 
    { 0x00002600, 0x01A0C600 }, 
    { 0x00002680, 0x01A0D680 }, 
    { 0x00002700, 0x01A0E700 }, 
    { 0x00002780, 0x01A0F780 }, 
    { 0x00002800, 0x01A10800 }, 
    { 0x00002880, 0x01A11880 }, 
    { 0x00002900, 0x01A12900 }, 
    { 0x00002980, 0x01A13980 }, 
    { 0x00002A00, 0x01A14A00 }, 
    { 0x00002A80, 0x01A15A80 }, 
    { 0x00002B00, 0x01A16B00 }, 
    { 0x00002B80, 0x01A17B80 }, 
    { 0x00002C00, 0x01A18C00 }, 
    { 0x00002C80, 0x01A19C80 }, 
    { 0x00002D00, 0x01A1AD00 }, 
    { 0x00002D80, 0x01A1BD80 }, 
    { 0x00002E00, 0x01A1CE00 }, 
    { 0x00002E80, 0x01A1DE80 }, 
    { 0x00002F00, 0x01A1EF00 }, 
    { 0x00002F80, 0x01A1FF80 }, 
    { 0x00003000, 0x01A20000 }, 
    { 0x00003080, 0x01A21080 }, 
    { 0x00003100, 0x01A22100 }, 
    { 0x00003180, 0x01A23180 }, 
    { 0x00003200, 0x01A24200 }, 
    { 0x00003280, 0x01A25280 }, 
    { 0x00003300, 0x01A26300 }, 
    { 0x00003380, 0x01A27380 }, 
    { 0x00003400, 0x01A28400 }, 
    { 0x00003480, 0x01A29480 }, 
    { 0x00003500, 0x01A2A500 }, 
    { 0x00003580, 0x01A2B580 }, 
    { 0x00003600, 0x01A2C600 }, 
    { 0x00003680, 0x01A2D680 }, 
    { 0x00003700, 0x01A2E700 }, 
    { 0x00003780, 0x01A2F780 }, 
    { 0x00003800, 0x01A30800 }, 
    { 0x00003880, 0x01A31880 }, 
    { 0x00003900, 0x01A32900 }, 
    { 0x00003980, 0x01A33980 }, 
    { 0x00003A00, 0x01A34A00 }, 
    { 0x00003A80, 0x01A35A80 }, 
    { 0x00003B00, 0x01A36B00 }, 
    { 0x00003B80, 0x01A37B80 }, 
    { 0x00003C00, 0x01A38C00 }, 
    { 0x00003C80, 0x01A39C80 }, 
    { 0x00003D00, 0x01A3AD00 }, 
    { 0x00003D80, 0x01A3BD80 }, 
    { 0x00003E00, 0x01A3CE00 }, 
    { 0x00003E80, 0x01A3DE80 }, 
    { 0x00003F00, 0x01A3EF00 }, 
    { 0x00003F80, 0x01A3FF80 }, 
    { 0x00004000, 0x01A40000 }, 
    { 0x00004080, 0x01A41080 }, 
    { 0x00004100, 0x01A42100 }, 
    { 0x00004180, 0x01A43180 }, 
    { 0x00004200, 0x01A44200 }, 
    { 0x00004280, 0x01A45280 }, 
    { 0x00004300, 0x01A46300 }, 
    { 0x00004380, 0x01A47380 }, 
    { 0x00004400, 0x01A48400 }, 
    { 0x00004480, 0x01A49480 }, 
    { 0x00004500, 0x01A4A500 }, 
    { 0x00004580, 0x01A4B580 }, 
    { 0x00004600, 0x01A4C600 }, 
    { 0x00004680, 0x01A4D680 }, 
    { 0x00004700, 0x01A4E700 }, 
    { 0x00004780, 0x01A4F780 }, 
    { 0x00004800, 0x01A50800 }, 
    { 0x00004880, 0x01A51880 }, 
    { 0x00004900, 0x01A52900 }, 
    { 0x00004980, 0x01A53980 }, 
    { 0x00004A00, 0x01A54A00 }, 
    { 0x00004A80, 0x01A55A80 }, 
    { 0x00004B00, 0x01A56B00 }, 
    { 0x00004B80, 0x01A57B80 }, 
    { 0x00004C00, 0x01A58C00 }, 
    { 0x00004C80, 0x01A59C80 }, 
    { 0x00004D00, 0x01A5AD00 }, 
    { 0x00004D80, 0x01A5BD80 }, 
    { 0x00004E00, 0x01A5CE00 }, 
    { 0x00004E80, 0x01A5DE80 }, 
    { 0x00004F00, 0x01A5EF00 }, 
    { 0x00004F80, 0x01A5FF80 }, 
    { 0x00005000, 0x01A60000 }, 
    { 0x00005080, 0x01A61080 }, 
    { 0x00005100, 0x01A62100 }, 
    { 0x00005180, 0x01A63180 }, 
    { 0x00005200, 0x01A64200 }, 
    { 0x00005280, 0x01A65280 }, 
    { 0x00005300, 0x01A66300 }, 
    { 0x00005380, 0x01A67380 }, 
    { 0x00005400, 0x01A68400 }, 
    { 0x00005480, 0x01A69480 }, 
    { 0x00005500, 0x01A6A500 }, 
    { 0x00005580, 0x01A6B580 }, 
    { 0x00005600, 0x01A6C600 }, 
    { 0x00005680, 0x01A6D680 }, 
    { 0x00005700, 0x01A6E700 }, 
    { 0x00005780, 0x01A6F780 }, 
    { 0x00005800, 0x01A70800 }, 
    { 0x00005880, 0x01A71880 }, 
    { 0x00005900, 0x01A72900 }, 
    { 0x00005980, 0x01A73980 }, 
    { 0x00005A00, 0x01A74A00 }, 
    { 0x00005A80, 0x01A75A80 }, 
    { 0x00005B00, 0x01A76B00 }, 
    { 0x00005B80, 0x01A77B80 }, 
    { 0x00005C00, 0x01A78C00 }, 
    { 0x00005C80, 0x01A79C80 }, 
    { 0x00005D00, 0x01A7AD00 }, 
    { 0x00005D80, 0x01A7BD80 }, 
    { 0x00005E00, 0x01A7CE00 }, 
    { 0x00005E80, 0x01A7DE80 }, 
    { 0x00005F00, 0x01A7EF00 }, 
    { 0x00005F80, 0x01A7FF80 }, 
    { 0x00006000, 0x01A80000 }, 
    { 0x00006080, 0x01A81080 }, 
    { 0x00006100, 0x01A82100 }, 
    { 0x00006180, 0x01A83180 }, 
    { 0x00006200, 0x01A84200 }, 
    { 0x00006280, 0x01A85280 }, 
    { 0x00006300, 0x01A86300 }, 
    { 0x00006380, 0x01A87380 }, 
    { 0x00006400, 0x01A88400 }, 
    { 0x00006480, 0x01A89480 }, 
    { 0x00006500, 0x01A8A500 }, 
    { 0x00006580, 0x01A8B580 }, 
    { 0x00006600, 0x01A8C600 }, 
    { 0x00006680, 0x01A8D680 }, 
    { 0x00006700, 0x01A8E700 }, 
    { 0x00006780, 0x01A8F780 }, 
    { 0x00006800, 0x01A90800 }, 
    { 0x00006880, 0x01A91880 }, 
    { 0x00006900, 0x01A92900 }, 
    { 0x00006980, 0x01A93980 }, 
    { 0x00006A00, 0x01A94A00 }, 
    { 0x00006A80, 0x01A95A80 }, 
    { 0x00006B00, 0x01A96B00 }, 
    { 0x00006B80, 0x01A97B80 }, 
    { 0x00006C00, 0x01A98C00 }, 
    { 0x00006C80, 0x01A99C80 }, 
    { 0x00006D00, 0x01A9AD00 }, 
    { 0x00006D80, 0x01A9BD80 }, 
    { 0x00006E00, 0x01A9CE00 }, 
    { 0x00006E80, 0x01A9DE80 }, 
    { 0x00006F00, 0x01A9EF00 }, 
    { 0x00006F80, 0x01A9FF80 }, 
    { 0x00007000, 0x01AA0000 }, 
    { 0x00007080, 0x01AA1080 }, 
    { 0x00007100, 0x01AA2100 }, 
    { 0x00007180, 0x01AA3180 }, 
    { 0x00007200, 0x01AA4200 }, 
    { 0x00007280, 0x01AA5280 }, 
    { 0x00007300, 0x01AA6300 }, 
    { 0x00007380, 0x01AA7380 }, 
    { 0x00007400, 0x01AA8400 }, 
    { 0x00007480, 0x01AA9480 }, 
    { 0x00007500, 0x01AAA500 }, 
    { 0x00007580, 0x01AAB580 }, 
    { 0x00007600, 0x01AAC600 }, 
    { 0x00007680, 0x01AAD680 }, 
    { 0x00007700, 0x01AAE700 }, 
    { 0x00007780, 0x01AAF780 }, 
    { 0x00007800, 0x01AB0800 }, 
    { 0x00007880, 0x01AB1880 }, 
    { 0x00007900, 0x01AB2900 }, 
    { 0x00007980, 0x01AB3980 }, 
    { 0x00007A00, 0x01AB4A00 }, 
    { 0x00007A80, 0x01AB5A80 }, 
    { 0x00007B00, 0x01AB6B00 }, 
    { 0x00007B80, 0x01AB7B80 }, 
    { 0x00007C00, 0x01AB8C00 }, 
    { 0x00007C80, 0x01AB9C80 }, 
    { 0x00007D00, 0x01ABAD00 }, 
    { 0x00007D80, 0x01ABBD80 }, 
    { 0x00007E00, 0x01ABCE00 }, 
    { 0x00007E80, 0x01ABDE80 }, 
    { 0x00007F00, 0x01ABEF00 }, 
    { 0x00007F80, 0x01ABFF80 }, 
    { 0x00008000, 0x01AC0000 }, 
    { 0x00008080, 0x01AC1080 }, 
    { 0x00008100, 0x01AC2100 }, 
    { 0x00008180, 0x01AC3180 }, 
    { 0x00008200, 0x01AC4200 }, 
    { 0x00008280, 0x01AC5280 }, 
    { 0x00008300, 0x01AC6300 }, 
    { 0x00008380, 0x01AC7380 }, 
    { 0x00008400, 0x01AC8400 }, 
    { 0x00008480, 0x01AC9480 }, 
    { 0x00008500, 0x01ACA500 }, 
    { 0x00008580, 0x01ACB580 }, 
    { 0x00008600, 0x01ACC600 }, 
    { 0x00008680, 0x01ACD680 }, 
    { 0x00008700, 0x01ACE700 }, 
    { 0x00008780, 0x01ACF780 }, 
    { 0x00008800, 0x01AD0800 }, 
    { 0x00008880, 0x01AD1880 }, 
    { 0x00008900, 0x01AD2900 }, 
    { 0x00008980, 0x01AD3980 }, 
    { 0x00008A00, 0x01AD4A00 }, 
    { 0x00008A80, 0x01AD5A80 }, 
    { 0x00008B00, 0x01AD6B00 }, 
    { 0x00008B80, 0x01AD7B80 }, 
    { 0x00008C00, 0x01AD8C00 }, 
    { 0x00008C80, 0x01AD9C80 }, 
    { 0x00008D00, 0x01ADAD00 }, 
    { 0x00008D80, 0x01ADBD80 }, 
    { 0x00008E00, 0x01ADCE00 }, 
    { 0x00008E80, 0x01ADDE80 }, 
    { 0x00008F00, 0x01ADEF00 }, 
    { 0x00008F80, 0x01ADFF80 }, 
    { 0x00009000, 0x01AE0000 }, 
    { 0x00009080, 0x01AE1080 }, 
    { 0x00009100, 0x01AE2100 }, 
    { 0x00009180, 0x01AE3180 }, 
    { 0x00009200, 0x01AE4200 }, 
    { 0x00009280, 0x01AE5280 }, 
    { 0x00009300, 0x01AE6300 }, 
    { 0x00009380, 0x01AE7380 }, 
    { 0x00009400, 0x01AE8400 }, 
    { 0x00009480, 0x01AE9480 }, 
    { 0x00009500, 0x01AEA500 }, 
    { 0x00009580, 0x01AEB580 }, 
    { 0x00009600, 0x01AEC600 }, 
    { 0x00009680, 0x01AED680 }, 
    { 0x00009700, 0x01AEE700 }, 
    { 0x00009780, 0x01AEF780 }, 
    { 0x00009800, 0x01AF0800 }, 
    { 0x00009880, 0x01AF1880 }, 
    { 0x00009900, 0x01AF2900 }, 
    { 0x00009980, 0x01AF3980 }, 
    { 0x00009A00, 0x01AF4A00 }, 
    { 0x00009A80, 0x01AF5A80 }, 
    { 0x00009B00, 0x01AF6B00 }, 
    { 0x00009B80, 0x01AF7B80 }, 
    { 0x00009C00, 0x01AF8C00 }, 
    { 0x00009C80, 0x01AF9C80 }, 
    { 0x00009D00, 0x01AFAD00 }, 
    { 0x00009D80, 0x01AFBD80 }, 
    { 0x00009E00, 0x01AFCE00 }, 
    { 0x00009E80, 0x01AFDE80 }, 
    { 0x00009F00, 0x01AFEF00 }, 
    { 0x00009F80, 0x01AFFF80 }, 
    { 0x0000A000, 0x01B00000 }, 
    { 0x0000A080, 0x01B01080 }, 
    { 0x0000A100, 0x01B02100 }, 
    { 0x0000A180, 0x01B03180 }, 
    { 0x0000A200, 0x01B04200 }, 
    { 0x0000A280, 0x01B05280 }, 
    { 0x0000A300, 0x01B06300 }, 
    { 0x0000A380, 0x01B07380 }, 
    { 0x0000A400, 0x01B08400 }, 
    { 0x0000A480, 0x01B09480 }, 
    { 0x0000A500, 0x01B0A500 }, 
    { 0x0000A580, 0x01B0B580 }, 
    { 0x0000A600, 0x01B0C600 }, 
    { 0x0000A680, 0x01B0D680 }, 
    { 0x0000A700, 0x01B0E700 }, 
    { 0x0000A780, 0x01B0F780 }, 
    { 0x0000A800, 0x01B10800 }, 
    { 0x0000A880, 0x01B11880 }, 
    { 0x0000A900, 0x01B12900 }, 
    { 0x0000A980, 0x01B13980 }, 
    { 0x0000AA00, 0x01B14A00 }, 
    { 0x0000AA80, 0x01B15A80 }, 
    { 0x0000AB00, 0x01B16B00 }, 
    { 0x0000AB80, 0x01B17B80 }, 
    { 0x0000AC00, 0x01B18C00 }, 
    { 0x0000AC80, 0x01B19C80 }, 
    { 0x0000AD00, 0x01B1AD00 }, 
    { 0x0000AD80, 0x01B1BD80 }, 
    { 0x0000AE00, 0x01B1CE00 }, 
    { 0x0000AE80, 0x01B1DE80 }, 
    { 0x0000AF00, 0x01B1EF00 }, 
    { 0x0000AF80, 0x01B1FF80 }, 
    { 0x0000B000, 0x01B20000 }, 
    { 0x0000B080, 0x01B21080 }, 
    { 0x0000B100, 0x01B22100 }, 
    { 0x0000B180, 0x01B23180 }, 
    { 0x0000B200, 0x01B24200 }, 
    { 0x0000B280, 0x01B25280 }, 
    { 0x0000B300, 0x01B26300 }, 
    { 0x0000B380, 0x01B27380 }, 
    { 0x0000B400, 0x01B28400 }, 
    { 0x0000B480, 0x01B29480 }, 
    { 0x0000B500, 0x01B2A500 }, 
    { 0x0000B580, 0x01B2B580 }, 
    { 0x0000B600, 0x01B2C600 }, 
    { 0x0000B680, 0x01B2D680 }, 
    { 0x0000B700, 0x01B2E700 }, 
    { 0x0000B780, 0x01B2F780 }, 
    { 0x0000B800, 0x01B30800 }, 
    { 0x0000B880, 0x01B31880 }, 
    { 0x0000B900, 0x01B32900 }, 
    { 0x0000B980, 0x01B33980 }, 
    { 0x0000BA00, 0x01B34A00 }, 
    { 0x0000BA80, 0x01B35A80 }, 
    { 0x0000BB00, 0x01B36B00 }, 
    { 0x0000BB80, 0x01B37B80 }, 
    { 0x0000BC00, 0x01B38C00 }, 
    { 0x0000BC80, 0x01B39C80 }, 
    { 0x0000BD00, 0x01B3AD00 }, 
    { 0x0000BD80, 0x01B3BD80 }, 
    { 0x0000BE00, 0x01B3CE00 }, 
    { 0x0000BE80, 0x01B3DE80 }, 
    { 0x0000BF00, 0x01B3EF00 }, 
    { 0x0000BF80, 0x01B3FF80 }, 
    { 0x0000C000, 0x01B40000 }, 
    { 0x0000C080, 0x01B41080 }, 
    { 0x0000C100, 0x01B42100 }, 
    { 0x0000C180, 0x01B43180 }, 
    { 0x0000C200, 0x01B44200 }, 
    { 0x0000C280, 0x01B45280 }, 
    { 0x0000C300, 0x01B46300 }, 
    { 0x0000C380, 0x01B47380 }, 
    { 0x0000C400, 0x01B48400 }, 
    { 0x0000C480, 0x01B49480 }, 
    { 0x0000C500, 0x01B4A500 }, 
    { 0x0000C580, 0x01B4B580 }, 
    { 0x0000C600, 0x01B4C600 }, 
    { 0x0000C680, 0x01B4D680 }, 
    { 0x0000C700, 0x01B4E700 }, 
    { 0x0000C780, 0x01B4F780 }, 
    { 0x0000C800, 0x01B50800 }, 
    { 0x0000C880, 0x01B51880 }, 
    { 0x0000C900, 0x01B52900 }, 
    { 0x0000C980, 0x01B53980 }, 
    { 0x0000CA00, 0x01B54A00 }, 
    { 0x0000CA80, 0x01B55A80 }, 
    { 0x0000CB00, 0x01B56B00 }, 
    { 0x0000CB80, 0x01B57B80 }, 
    { 0x0000CC00, 0x01B58C00 }, 
    { 0x0000CC80, 0x01B59C80 }, 
    { 0x0000CD00, 0x01B5AD00 }, 
    { 0x0000CD80, 0x01B5BD80 }, 
    { 0x0000CE00, 0x01B5CE00 }, 
    { 0x0000CE80, 0x01B5DE80 }, 
    { 0x0000CF00, 0x01B5EF00 }, 
    { 0x0000CF80, 0x01B5FF80 }, 
    { 0x0000D000, 0x01B60000 }, 
    { 0x0000D080, 0x01B61080 }, 
    { 0x0000D100, 0x01B62100 }, 
    { 0x0000D180, 0x01B63180 }, 
    { 0x0000D200, 0x01B64200 }, 
    { 0x0000D280, 0x01B65280 }, 
    { 0x0000D300, 0x01B66300 }, 
    { 0x0000D380, 0x01B67380 }, 
    { 0x0000D400, 0x01B68400 }, 
    { 0x0000D480, 0x01B69480 }, 
    { 0x0000D500, 0x01B6A500 }, 
    { 0x0000D580, 0x01B6B580 }, 
    { 0x0000D600, 0x01B6C600 }, 
    { 0x0000D680, 0x01B6D680 }, 
    { 0x0000D700, 0x01B6E700 }, 
    { 0x0000D780, 0x01B6F780 }, 
    { 0x0000D800, 0x01B70800 }, 
    { 0x0000D880, 0x01B71880 }, 
    { 0x0000D900, 0x01B72900 }, 
    { 0x0000D980, 0x01B73980 }, 
    { 0x0000DA00, 0x01B74A00 }, 
    { 0x0000DA80, 0x01B75A80 }, 
    { 0x0000DB00, 0x01B76B00 }, 
    { 0x0000DB80, 0x01B77B80 }, 
    { 0x0000DC00, 0x01B78C00 }, 
    { 0x0000DC80, 0x01B79C80 }, 
    { 0x0000DD00, 0x01B7AD00 }, 
    { 0x0000DD80, 0x01B7BD80 }, 
    { 0x0000DE00, 0x01B7CE00 }, 
    { 0x0000DE80, 0x01B7DE80 }, 
    { 0x0000DF00, 0x01B7EF00 }, 
    { 0x0000DF80, 0x01B7FF80 }, 
    { 0x0000E000, 0x01B80000 }, 
    { 0x0000E080, 0x01B81080 }, 
    { 0x0000E100, 0x01B82100 }, 
    { 0x0000E180, 0x01B83180 }, 
    { 0x0000E200, 0x01B84200 }, 
    { 0x0000E280, 0x01B85280 }, 
    { 0x0000E300, 0x01B86300 }, 
    { 0x0000E380, 0x01B87380 }, 
    { 0x0000E400, 0x01B88400 }, 
    { 0x0000E480, 0x01B89480 }, 
    { 0x0000E500, 0x01B8A500 }, 
    { 0x0000E580, 0x01B8B580 }, 
    { 0x0000E600, 0x01B8C600 }, 
    { 0x0000E680, 0x01B8D680 }, 
    { 0x0000E700, 0x01B8E700 }, 
    { 0x0000E780, 0x01B8F780 }, 
    { 0x0000E800, 0x01B90800 }, 
    { 0x0000E880, 0x01B91880 }, 
    { 0x0000E900, 0x01B92900 }, 
    { 0x0000E980, 0x01B93980 }, 
    { 0x0000EA00, 0x01B94A00 }, 
    { 0x0000EA80, 0x01B95A80 }, 
    { 0x0000EB00, 0x01B96B00 }, 
    { 0x0000EB80, 0x01B97B80 }, 
    { 0x0000EC00, 0x01B98C00 }, 
    { 0x0000EC80, 0x01B99C80 }, 
    { 0x0000ED00, 0x01B9AD00 }, 
    { 0x0000ED80, 0x01B9BD80 }, 
    { 0x0000EE00, 0x01B9CE00 }, 
    { 0x0000EE80, 0x01B9DE80 }, 
    { 0x0000EF00, 0x01B9EF00 }, 
    { 0x0000EF80, 0x01B9FF80 }, 
    { 0x0000F000, 0x01BA0000 }, 
    { 0x0000F080, 0x01BA1080 }, 
    { 0x0000F100, 0x01BA2100 }, 
    { 0x0000F180, 0x01BA3180 }, 
    { 0x0000F200, 0x01BA4200 }, 
    { 0x0000F280, 0x01BA5280 }, 
    { 0x0000F300, 0x01BA6300 }, 
    { 0x0000F380, 0x01BA7380 }, 
    { 0x0000F400, 0x01BA8400 }, 
    { 0x0000F480, 0x01BA9480 }, 
    { 0x0000F500, 0x01BAA500 }, 
    { 0x0000F580, 0x01BAB580 }, 
    { 0x0000F600, 0x01BAC600 }, 
    { 0x0000F680, 0x01BAD680 }, 
    { 0x0000F700, 0x01BAE700 }, 
    { 0x0000F780, 0x01BAF780 }, 
    { 0x0000F800, 0x01BB0800 }, 
    { 0x0000F880, 0x01BB1880 }, 
    { 0x0000F900, 0x01BB2900 }, 
    { 0x0000F980, 0x01BB3980 }, 
    { 0x0000FA00, 0x01BB4A00 }, 
    { 0x0000FA80, 0x01BB5A80 }, 
    { 0x0000FB00, 0x01BB6B00 }, 
    { 0x0000FB80, 0x01BB7B80 }, 
    { 0x0000FC00, 0x01BB8C00 }, 
    { 0x0000FC80, 0x01BB9C80 }, 
    { 0x0000FD00, 0x01BBAD00 }, 
    { 0x0000FD80, 0x01BBBD80 }, 
    { 0x0000FE00, 0x01BBCE00 }, 
    { 0x0000FE80, 0x01BBDE80 }, 
    { 0x0000FF00, 0x01BBEF00 }, 
    { 0x0000FF80, 0x01BBFF80 }, 
    { 0x00010000, 0x01BC0000 }, 
    { 0x00010080, 0x01BC1080 }, 
    { 0x00010100, 0x01BC2100 }, 
    { 0x00010180, 0x01BC3180 }, 
    { 0x00010200, 0x01BC4200 }, 
    { 0x00010280, 0x01BC5280 }, 
    { 0x00010300, 0x01BC6300 }, 
    { 0x00010380, 0x01BC7380 }, 
    { 0x00010400, 0x01BC8400 }, 
    { 0x00010480, 0x01BC9480 }, 
    { 0x00010500, 0x01BCA500 }, 
    { 0x00010580, 0x01BCB580 }, 
    { 0x00010600, 0x01BCC600 }, 
    { 0x00010680, 0x01BCD680 }, 
    { 0x00010700, 0x01BCE700 }, 
    { 0x00010780, 0x01BCF780 }, 
    { 0x00010800, 0x01BD0800 }, 
    { 0x00010880, 0x01BD1880 }, 
    { 0x00010900, 0x01BD2900 }, 
    { 0x00010980, 0x01BD3980 }, 
    { 0x00010A00, 0x01BD4A00 }, 
    { 0x00010A80, 0x01BD5A80 }, 
    { 0x00010B00, 0x01BD6B00 }, 
    { 0x00010B80, 0x01BD7B80 }, 
    { 0x00010C00, 0x01BD8C00 }, 
    { 0x00010C80, 0x01BD9C80 }, 
    { 0x00010D00, 0x01BDAD00 }, 
    { 0x00010D80, 0x01BDBD80 }, 
    { 0x00010E00, 0x01BDCE00 }, 
    { 0x00010E80, 0x01BDDE80 }, 
    { 0x00010F00, 0x01BDEF00 }, 
    { 0x00010F80, 0x01BDFF80 }, 
    { 0x00011000, 0x01BE0000 }, 
    { 0x00011080, 0x01BE1080 }, 
    { 0x00011100, 0x01BE2100 }, 
    { 0x00011180, 0x01BE3180 }, 
    { 0x00011200, 0x01BE4200 }, 
    { 0x00011280, 0x01BE5280 }, 
    { 0x00011300, 0x01BE6300 }, 
    { 0x00011380, 0x01BE7380 }, 
    { 0x00011400, 0x01BE8400 }, 
    { 0x00011480, 0x01BE9480 }, 
    { 0x00011500, 0x01BEA500 }, 
    { 0x00011580, 0x01BEB580 }, 
    { 0x00011600, 0x01BEC600 }, 
    { 0x00011680, 0x01BED680 }, 
    { 0x00011700, 0x01BEE700 }, 
    { 0x00011780, 0x01BEF780 }, 
    { 0x00011800, 0x01BF0800 }, 
    { 0x00011880, 0x01BF1880 }, 
    { 0x00011900, 0x01BF2900 }, 
    { 0x00011980, 0x01BF3980 }, 
    { 0x00011A00, 0x01BF4A00 }, 
    { 0x00011A80, 0x01BF5A80 }, 
    { 0x00011B00, 0x01BF6B00 }, 
    { 0x00011B80, 0x01BF7B80 }, 
    { 0x00011C00, 0x01BF8C00 }, 
    { 0x00011C80, 0x01BF9C80 }, 
    { 0x00011D00, 0x01BFAD00 }, 
    { 0x00011D80, 0x01BFBD80 }, 
    { 0x00011E00, 0x01BFCE00 }, 
    { 0x00011E80, 0x01BFDE80 }, 
    { 0x00011F00, 0x01BFEF00 }, 
    { 0x00011F80, 0x01BFFF80 }, 
    { 0x00012000, 0x01C00000 }, 
    { 0x00012080, 0x01C01080 }, 
    { 0x00012100, 0x01C02100 }, 
    { 0x00012180, 0x01C03180 }, 
    { 0x00012200, 0x01C04200 }, 
    { 0x00012280, 0x01C05280 }, 
    { 0x00012300, 0x01C06300 }, 
    { 0x00012380, 0x01C07380 }, 
    { 0x00012400, 0x01C08400 }, 
    { 0x00012480, 0x01C09480 }, 
    { 0x00012500, 0x01C0A500 }, 
    { 0x00012580, 0x01C0B580 }, 
    { 0x00012600, 0x01C0C600 }, 
    { 0x00012680, 0x01C0D680 }, 
    { 0x00012700, 0x01C0E700 }, 
    { 0x00012780, 0x01C0F780 }, 
    { 0x00012800, 0x01C10800 }, 
    { 0x0001285C, 0x0001285C }, 
    { 0x00013000, 0x01800000 }, 
    { 0x00015E0C, 0x00015E0C }
};

// for MP2KUS only (needs to be verified for others) 
//   splitting code segment to the number of subsegments
static SegmentData code_seg_split[] = {
    // reset code
    { 0x00000000, false, "CODE",  "Boot Code" },
    // boot data??
    { 0x00000020, true,  "CONST", "Boot Data (Read-Only)" },
    // patch table 
    { 0x00002000, true,  "CONST", "ROM Patch Table (Read-Only)" },
    // jump table
    { 0x00013000, true,  "CONST", "ROM Public Jump Table (Read-Only)" },
    // patch table page data
    { 0x00016000, true,  "CONST", "ROM Patch Table Page Table (Read-Only)" },
    // jump table page data
    { 0x00018000, true,  "CONST", "ROM Public Jump Table Page Table (Read-Only)" },
    // first code segment
    { 0x00018400, false, "CODE",  "Code Segment 1" },
    // ROM static RO data
    { 0x0036340C, true,  "CONST", "ROM Data (Read-Only)" },
    // second code segment
    { 0x00382324, false, "CODE",  "Code Segment 2" },
    // magic pointer table
    { 0x003AF000, true,  "CONST", "ROM Magic Pointer Table (Read-Only)" },
    // ROM Soups (symbol data)
    { 0x003AFDA8, true,  "CONST", "ROM Soup Data (Read-Only)" }
};
 

// Limits for patch table
static uint32 ROMPatchTable = 0x00002000;
static uint32 ROMPatchTableEnd = 0x0001285C;
static uint32 ROMPublicJumpTable = 0x00013000;
static uint32 ROMPublicJumpTableEnd = 0x00015E0C;


// ------------------------ Functions ------------------------ 

// checks whether given address contains a data
bool is_data_address (uint32 value) {

    // data segments
    if (value >= data_base && value < (data_base + data_length) 
        || value >= zero_data_base && value < (zero_data_base + zero_data_length)) {
        
        return true;
    }
    
    // code segment
    if (value >= code_base && value < (code_base + code_length) && isMP2KUS) {
        // we have s specific split of the code segment for
        // MP2K US ROM - it has a number of data segments  
        // from there
        int num_segs = sizeof(code_seg_split)/sizeof(code_seg_split[0]);
        for (int i=0; i<num_segs; i++) {
            uint32 segm_base = code_seg_split[i].base;
            uint32 next_seg_base = code_base+code_length;
    
            // next segment base is this segment end
            if (i+1<num_segs) {
                next_seg_base = code_seg_split[i+1].base;
            }
    
            // check if address is in one of the segments
            if (value >= segm_base && value < next_seg_base) {
                return code_seg_split[i].is_data;
            }
        }
    }
    
    return false;
}

// checks if given address is in valid range (in one of the segments)
bool is_mapped_address (uint32 value) {
    bool result = false;

    // all mmu segments mapped from code segment
    for (int i=0; i<sizeof(mmu)/sizeof(mmu[0])-1; i++) {
        if (mmu[i].base == mmu[i].mapped_to) {
            continue;    // skip equal mappings (they are placeholders)
        }

        // calculate file offset and size
        uint32 length = mmu[i+1].base - mmu[i].base;

        result = result || 
                (value >= mmu[i].mapped_to && value < (mmu[i].mapped_to+length));

    }

    return result;
}

// checks if given address is in valid range (in one of the segments)
bool is_valid_address (uint32 value) {
    return (value >= code_base && value < (code_base + code_length)) 
           || (value >= data_base && value < (data_base + data_length)) 
           || (value >= zero_data_base && value < (zero_data_base + zero_data_length))
           || is_mapped_address (value);
}

// retrieves word32 from given address with given endianness
uint32 get_word32 (byte* data) {
    uint32 value = 0;
    
    if (endian == ET_BIG) {
        value = (data[0] << 24) | (data[1] << 16) | (data[2] << 8) | data[3];
    }
    else {
        value = (data[3] << 24) | (data[2] << 16) | (data[1] << 8) | data[0];
    }
    
    return value;
}

// retrieves word16 from given address with given endianness
uint16 get_word16 (byte* data) {
    uint16 value = 0;
    
    if (endian == ET_BIG) {
        value = (data[0] << 8) | data[1];
    }
    else {
        value = (data[1] << 8) | data[0];
    }
    
    return value;
}

// recognising the endian
void recognise_endian () {
    if (memcmp (header.program_exit, 
                prog_exit_big_endian, 
                sizeof (header.entry_point)) == 0)
    {
        endian = ET_BIG;
    }
    else if (memcmp (header.program_exit, 
                     prog_exit_little_endian, 
                     sizeof (header.entry_point)) == 0) 
    {
        endian = ET_LITTLE;
    }
    else {
        printf ("WARNING: cant's deduce byte endian-ness - using big-endian by default!\n");
        endian = ET_BIG;
    }
}

// retrieves name by given offset in a string table
char* get_name(uint32 offset) {
    StringTable * strings = (StringTable *) string_table;
    
    if (offset < get_word32(strings->length))
    {
        StringEntry * str = (StringEntry *) (string_table + offset);
        return str->str_data;
    }
    
    return "INVALID OFFSET!";
}

// process symbol
void process_symbol (Symbol * symbol) {
    uint8 flags = (uint16)(get_word32(symbol->symbol) >> 24) & 0xF;
    char * orig_name = get_name (get_word32(symbol->symbol) & 0xFFFFFF);
    uint32 value = get_word32(symbol->value);
    uint8 is_global = flags & ASD_ABSSYM;
    char name [500];
    bool do_free = false;
    char str[500];
    char short_name_buf [500];
    char * short_name = short_name_buf;
    char * demangled_name = 0;
    char * pos = 0;

    // demangle name
    demangled_name = cplus_demangle (orig_name, DMGL_PARAMS|DMGL_ANSI|DMGL_TYPES);
    if (demangled_name) {
        strcpy (name, demangled_name);
        free (demangled_name);
    }
    else {
        strcpy (name, orig_name);
    }

    // try to set type and name
    if (is_valid_address(value)) {
        // try to get short name
        if (truncate_function_name  && (pos=strchr(name, '('))) {
            strncpy (short_name_buf+1, name, pos-name);
            short_name_buf[pos-name+1] = '\0';
        }
        else {
            strcpy (short_name_buf+1, name);
        }
        // check the name
        if (is_mapped_address (value)) {
            // add another character to distinguish it from keyword
            short_name_buf[0] = '@';
        }
        else if (validate_name (short_name_buf+1)) {
            short_name++;
        }
        else {
            // add another character to distinguish it from keyword
            short_name_buf[0] = '$';
        }
        
        // try to set the name
        if (!do_name_anyway(value, short_name)) {
            sprintf (str, "Cannot set name \"%s\" for the address 0x%08X\n", name, value);
            msg (str);
        }
    
        // try also to set a comment - full name
        set_cmt (value, name, false);
    }
}

// process symbol table
void process_symbol_table (uint32 nsyms) {
    msg ("Loading symbols...");
    
    // set demangling style
    cplus_demangle_set_style (arm_demangling);
    
    for (uint32 i=0; i<nsyms; i++) {
        process_symbol (&symbols[i]);
    }
    msg ("done!\n");
}

// create single segment
bool add_segment (FILE* fi, uint32 file_offset, uint32 start, uint32 end, 
                  const char *name, const char *seg_class) {
    segment_t * segment = 0;
    
    // try to create
    if (!add_segm(1, start, end, name, seg_class)) {
        return false;
    }
    
    // get greated segment
    segment = getseg (start);
    
    // alter default segment data
    segment->bitness = 1;        // 32-bit
    segment->align = saRelDble;  // 4 bytes alignment
    if (!segment->update()) {
        return false;
    }

    // and get the content from the file (if this is not a memory based segment)
    if (fi) {
        file2base(fi, file_offset, start, end, FILEREG_PATCHABLE);
    }

    return true;
}

// add and populate code segment
bool add_code_segment (FILE* fi) {
    if (!isMP2KUS) {
        msg ("Code segment:\n");
        if (!add_segment(fi, code_offset, code_base, code_base+code_length-1, 
                         "Code segment", "CODE")) {
            return false;
        }
    }
    else {
        // we have s specific split of the code segment for
        // MP2K US ROM - use a number of code/data segments  
        // from there
        msg ("The image recognised as MP2x00 US Debug ROM - splitting code segment:\n");
        int num_segs = sizeof(code_seg_split)/sizeof(code_seg_split[0]);
        for (int i=0; i<num_segs; i++) {
            uint32 segm_base = code_seg_split[i].base;
            uint32 segm_offset = code_offset+segm_base;
            uint32 next_seg_base = code_base+code_length;

            // next segment base is this segment end
            if (i+1<num_segs) {
                next_seg_base = code_seg_split[i+1].base;
            }

            // try to add segment
            if (!add_segment(fi, segm_offset, segm_base, next_seg_base, 
                             code_seg_split[i].name, code_seg_split[i].type)) {
                return false;
            }
        }
    }

    return true;
}


// adding MMU segments
bool add_memory_mapped_segments (FILE* fi) {
    const char * segment_name = "Mapped ROM Code";

    // all mmu segments mapped from code segment
    for (int i=0; i<sizeof(mmu)/sizeof(mmu[0])-1; i++) {
        if (mmu[i].base == mmu[i].mapped_to) {
            continue;    // skip equal mappings (they are placeholders)
        }
        
        // calculate file offset and size
        uint32 offset = code_offset + mmu[i].base - code_base;
        uint32 base = mmu[i].mapped_to;
        uint32 length = mmu[i+1].base - mmu[i].base;

        // choose the name
        if (mmu[i].base >= ROMPatchTable 
            && mmu[i].base < ROMPatchTableEnd) 
        {
            segment_name = "Mapped ROM Patch Table";
        }
        else if (mmu[i].base >= ROMPublicJumpTable 
                 && mmu[i].base < ROMPublicJumpTableEnd) 
        {
            segment_name = "Mapped ROM Public Jump Table";
        }

        // try to add the segment
        if (!add_segment(fi, offset, base, base+length, segment_name, "CODE")) {
            return false;
        }
    }

    return true;
}

// add and populate segments
bool add_segments (FILE* fi) {
    char* buf = 0;

    // map selector 1 to 0 - this is ridiculous for 
    // non intel/PC architecture but IDA requires it
    set_selector(1, 0);

    // 1. create code segment
    if (!add_code_segment (fi)) {
        return false;
    }

    // 2. create data segment
    msg ("Read-write data segment:\n");
    if (!add_segment(fi, data_offset, data_base, data_base+data_length, 
                     "Initialised Data segment (Read/Write)", "DATA")) {
        return false;
    }

    // 3. create zero-init data segment
    msg ("Zero-initialised read-write data segment:\n");
    if (!add_segment(0, 0, zero_data_base, zero_data_base+zero_data_length, 
                     "Zero Initialised Data segment (Read/Write)", "DATA")) {
        return false;
    }

    // allocate some data in a buffer filled with 0
    if (!(buf = (char *) qcalloc (zero_data_length, 1))){
        msg ("Cannot allocate enough memory for Zero Data!");
        return false;
    }
    // fill in zero-init data from memory
    mem2base(buf, zero_data_base, zero_data_base+zero_data_length, -1);
    
    // clean-up
    qfree (buf);
    buf = 0;

    // do MMU stuff
    msg ("Memory mapped segments:\n");
    return add_memory_mapped_segments (fi);
}

// read header and guess the endian
int read_header (FILE* fi) {
    // read the header
    qfseek(fi, 0, SEEK_SET);
    if (qfread(fi, (void*)&header, sizeof(header)) != sizeof(header)) 
        return 0;

    // recognise endianness
    recognise_endian ();
    
    code_offset = sizeof(header);
    code_base = get_word32 (header.image_base);
    code_length = get_word32 (header.ro_area_size);
    
    data_offset = code_offset + code_length;
    if (get_word32(header.addressing_type) & 0x100) {
        data_base = get_word32 (header.data_base);
    }
    else {
        data_base = code_base + code_length;
    }
    data_length = get_word32 (header.rw_area_size);;
    
    zero_data_base = data_base + data_length;
    zero_data_length = get_word32 (header.zero_init_area_size);
    
    debug_offset = data_offset + data_length;
    debug_length = get_word32(header.debug_area_size);

    return 1;
}

/*
 *   Check if input file can be a newton MP image file. If it passes return 
 *   and fill in the formatname otherwise return 0
 */

int accept_file(FILE *fi, char fileformatname[MAX_FILE_FORMAT_NAME], int n) {
    ulong file_length = 0;

    if (n) {
        return 0;
    }

    // first get the lenght of the file
    file_length = efilelength(fi);
    
    // if to short for a AIF header then this is no AIF
    if (file_length < sizeof(header)) {
        return 0;
    }

    // read AIF header from file and parse some data sizes from it
    if (read_header(fi) == 0) 
        return 0;
  
    // now perform some sanitychecks
    if (debug_offset + debug_length != file_length)
        return 0;
        
    // don't support 26 bit addressing mode
    if (get_word32(header.addressing_type) & 0xff != 32)
        return 0;
    
    // file has passed all sanity checks and might be a DOL
    qstrncpy(fileformatname, "Newton Message Pad ROM Image", MAX_FILE_FORMAT_NAME);
    
    // we need ARM support otherwise we cannot do much with Newton ROM
    if (ph.id != PLFM_ARM) {
        if (endian == ET_BIG) {
            set_processor_type("ARMB", SETPROC_ALL|SETPROC_FATAL);
        }
        else {
            set_processor_type("ARM", SETPROC_ALL|SETPROC_FATAL);
        }
    }

    return ACCEPT_FIRST | 1;
}

//
// Load the recognised file into the database.
//
void load_file(FILE *fi, ushort /*neflag*/, const char * /*fileformatname*/) {
    char* buf = 0;

    // message
    msg("-------------------------------------------\n");
    msg("Newton Message Pad ROM Image Loader ver.1.0\n");
    msg("-------------------------------------------\n");
  
    // we need ARM support otherwise we cannot do much with Newton ROM
    if (ph.id != PLFM_ARM) {
        if (endian == ET_BIG) {
            set_processor_type("ARMB", SETPROC_ALL|SETPROC_FATAL);
        }
        else {
            set_processor_type("ARM", SETPROC_ALL|SETPROC_FATAL);
        }
    }

    // check if it's an MP2K US Image mp2kus_image_size
    isMP2KUS = (efilelength(fi) == mp2kus_image_size);

    // read header again - the module may have been unloaded
    if (read_header(fi) == 0) {
        qexit(1);
    }
  
    // set entry points
    inf.beginEA = inf.startIP = get_word32 (header.entry_point);

    // add segmens and load data
    if (!add_segments (fi)) {
        qexit(1);
    }

    // process symbol information
    if (debug_length > 0) {
        if (buf = (char *) qalloc (debug_length)) {
            qfseek(fi, debug_offset, SEEK_SET);
            if ((uint32)qfread(fi, (void*)buf, debug_length) == debug_length) {
                // all data read successfully - set structures
                SectionItem * section = (SectionItem *) buf;
                int section_type = get_word32(section->length_type) & 0xFFFF;
                
                if (section_type == 1 && section->language == LANG_NONE)
                {
                    // set up global variables
                    symbols = (Symbol *) (buf + sizeof (SectionItem));
                    string_table = (byte *) buf + sizeof (SectionItem) + get_word32(section->name_nsyms)*sizeof(Symbol);
                    
                    // process all symbols
                    process_symbol_table(get_word32(section->name_nsyms));
                }
                else {
                    printf ("Unsupported section type (%d) or language (%d) for debug symbols - no symbols will be present!\n", 
                            section_type, section->language);
                }
            }
            else {
                msg ("Cannot read debug symbol data - no symbols will be present!\n");
            }
                
            // clean-up
            qfree (buf);
        }
        else {
            msg ("Cannot allocate enough memory for debug symbols - no symbols will be present!\n");
        }
    }
}


//
// Loader Module Descriptor Blocks
//
extern "C" loader_t LDSC = {
  IDP_INTERFACE_VERSION,
  0, /* no loader flags */
  accept_file,
  load_file,
  NULL,
};

